"Contains word lists used for spell checking in dictionary building."

import io
from pkg_resources import resource_stream
from typing import TextIO

def get_words_file(name: str) -> TextIO:
    """Get a text stream of line separated words.
    
    :param name: The name of the word list to retrieve. (words.txt or 
        extra_words.txt)
    :returns: A text stream.
    """

    return io.TextIOWrapper(resource_stream("grascii.words", name),
            encoding="utf-8")



"""Contains methods for detemining the similarity of grascii strokes."""

from typing import Tuple, Hashable, List, Set, cast, Dict, Iterable

equiv_nodes: Dict[str, Tuple[str, ...]] = {
    "S" : ("S", "Z"),
    "Z" : ("S", "Z"),
    "TD" : ("TD", "DT", "DD"),
    "DT" : ("TD", "DT", "DD"),
    "DD" : ("TD", "DT", "DD"),
    "MN" : ("MN", "MM"),
    "MM" : ("MN", "MM"),
    "TN" : ("TN", "DN"),
    "DN" : ("TN", "DN"),
    "TM" : ("TM", "DM"),
    "DM" : ("TM", "DM"),
    "NT" : ("NT", "ND"),
    "ND" : ("NT", "ND"),
    "MT" : ("MT", "MD"),
    "MD" : ("MT", "MD"),
    "DF" : ("DF", "DV", "TV"),
    "DV" : ("DF", "DV", "TV"),
    "TV" : ("DF", "DV", "TV"),
    "PNT" : ("PNT", "PND", "JNT", "JND"),
    "PND" : ("PNT", "PND", "JNT", "JND"),
    "JNT" : ("PNT", "PND", "JNT", "JND"),
    "JND" : ("PNT", "PND", "JNT", "JND"),
    }

def get_node(stroke: str) -> Tuple[str, ...]:
    return equiv_nodes.get(stroke, (stroke,))

disconnected_nodes: List[Hashable] = [
    ("O",),
    ("U",),
    get_node("PNT"),
    get_node("DF")
    ]

edges: List[Tuple[str, str]] = [
    ("K", "G"),
    ("R", "L"),
    ("L", "LD"),
    ("T", "D"),
    ("D", "DT"),
    ("I", "A"),
    ("A", "E"),
    ("NK", "N"),
    ("N", "M"),
    ("M", "MN"),
    ("N", "NG"),
    ("M", "NK"),
    ("NG", "NK"),
    ("M", "NG"), # add to graph
    ("NK", "MN"),
    ("SH", "S"),
    ("S", "SS"),
    ("S", "XS"),
    ("SS", "XS"),
    ("XS", "X"),
    ("X", "S"),
    ("S", "F"),
    ("F", "V"),
    ("SH", "CH"),
    ("F", "CH"),
    ("V", "J"),
    ("CH", "J"),
    ("CH", "P"),
    ("S", "P"),
    ("P", "B"),
    ("J", "B"),
    ("TH", "TN"),
    ("TN", "TM"),
    ("NT", "MT"),
    ("TH", "NT"),
    ]

trans_edges: List[Tuple[Hashable, Hashable]] = [(get_node(edge[0]), get_node(edge[1])) for edge in edges]

class SimilarityGraph():
    """Implements a basic graph with the capability of doing an n-layered
    breadth-first search."""

    def __init__(self):
        self.nodes = dict()

    def add_node(self, node: Hashable) -> None:
        """Add a node to the graph.
        
        :param node: An object that can be used as a dict key.
        """
        
        if node not in self.nodes:
            self.nodes[node] = set()

    def add_edge(self, edge: Tuple[Hashable, Hashable]) -> None:
        """Add an edge to the graph. If either node does not exist
        in the graph, it is added.
        
        :param edge: Two objects that can be used as a dict key.
        """

        assert len(edge) > 1
        self.add_node(edge[0])
        self.nodes[edge[0]].add(edge[1])
        self.add_node(edge[1])
        self.nodes[edge[1]].add(edge[0])

    def add_nodes(self, nodes: Iterable[Hashable]) -> None:
        """Add a collection of nodes to the graph.
        
        :param nodes: A collection of objects that can be used as a dict key.
        """

        for node in nodes:
            self.add_node(node)

    def add_edges(self, edges: Iterable[Tuple[Hashable, Hashable]]) -> None:
        """Add a collection of edges to the graph.
        
        :param nodes: A collection of edges.
        """

        for edge in edges:
            self.add_edge(edge)

    def get_similar(self, node: Hashable, similarity: int) -> Set[Hashable]:
        """Get a set of all nodes within a distance of similarity to the
        given node.

        :param node: The node to get similars for.
        :param similarity: How different... uncertainty.
        :returns: A set of nodes
        """

        similars = set()
        similars.add(node)
        while similarity > 0:
            new = list()
            for node in similars:
                new += self.nodes.get(node, list())
            similars |= set(new)
            similarity -= 1

        return similars

sg = SimilarityGraph()
sg.add_edges(trans_edges)
sg.add_nodes(disconnected_nodes)

def get_similar(stroke: str, distance: int) -> Set[Tuple]:
    """Get a set of all strokes within a distance to the
    given node.

    :param stroke: The stroke to get similars for.
    :param distance: How different... uncertainty.
    :returns: A set of strokes.
    """

    return cast(Set[Tuple], sg.get_similar(get_node(stroke), distance))





import io
import os
from typing import Union, IO, TextIO
from warnings import warn

from pkg_resources import resource_stream, resource_string

def get_grammar(name: str) -> str:
    warn("This function is deprecated. Use grascii.grammars.get_grammar instead", DeprecationWarning)
    return resource_string("grascii.grammars", name + ".lark").decode("utf-8")

# def get_dict_file(dictionary: str, name: str) -> TextIO:
    # if dictionary[0] == ":":
        # if dictionary == ":preanniversary":
            # return io.TextIOWrapper(resource_stream("grascii.dict", name),
                    # encoding="utf-8")
    # return open(os.path.join(dictionary, name))

def get_words_file(name: str) -> TextIO:
    warn("This function is deprecated. Use grascii.words.get_words_file instead", DeprecationWarning)
    return io.TextIOWrapper(resource_stream("grascii.words", name),
            encoding="utf-8")





"""
Contains RegexBuilder for generating regular expression to use in grascii
searches.
"""

from enum import Enum
import re
from typing import List, Union, Iterable, Set, Pattern, Tuple

from grascii import grammar
from grascii.similarities import get_similar
from grascii.types import Interpretation

class SearchMode(Enum):
    """An enum representing different searh modes."""

    MATCH = "match"
    START = "start"
    CONTAIN = "contain"

class Strictness(Enum):
    """An enum representing different levels of strictness for handling
    annotations, aspirates, and disjoiners in grascii strings."""

    LOW = "discard"
    MEDIUM = "retain"
    HIGH = "strict"

class RegexBuilder():
    """A class used for generating regular expressions used to search from
    a grascii string based on grascii search options.
    
    :param uncertainty: The uncertainty of the grascii string.
    :param search_mode: The search mode to use.
    :param annotation_mode: How to handle annotations in the search.
    :param aspirate_mode: How to handle annotations in the search.
    :param disjoiner_mode: How to handle annotations in the search.
    :param fix_first: Apply an uncertainty of 0 to the first token.
    :type uncertainty: int: 0, 1, or 2
    :type search_mode: str: one of regen.SearchMode values
    :type annotation_mode: one of regen.Strictness values
    :type aspirate_mode: one of regen.Strictness values
    :type disjoiner_mode: one of regen.Strictness values
    :type fix_first: bool
    """

    def __init__(self, **kwargs):
        self.uncertainty = kwargs.get('uncertainty', 0)
        self.search_mode = kwargs.get('search_mode', SearchMode.MATCH)
        self.fix_first = kwargs.get('fix_first', False)
        self.annotation_mode = kwargs.get('annotation_mode', Strictness.LOW)
        self.aspirate_mode = kwargs.get('aspirate_mode', Strictness.LOW)
        self.disjoiner_mode = kwargs.get('disjoiner_mode', Strictness.HIGH)

    def make_annotation_regex(self, stroke: str, annotations: Iterable[str]) -> str:
        """Create a regular expression that matches the stroke with
        the acceptable annotations and given annotations according to
        the search options.
        
        :param stroke: The stroke for which to generate annotations.
        :param annotations: A collection of annotations used in generating
            the regular expression.

        :returns: A regular expression.
        """

        def pack(tup):
            if len(tup) == 1:
                return re.escape(tup[0]) + "?"
            return "[" + "".join(tup) + "]?"

        possible = grammar.ANNOTATIONS.get(stroke, list())

        if self.annotation_mode is Strictness.MEDIUM or \
                self.annotation_mode is Strictness.HIGH:
            builder = list()
            i = 0
            while i < len(possible):
                match = False
                for an in annotations:
                    if an in possible[i]:
                        builder.append(re.escape(an))
                        match = True
                        break
                if not match and self.annotation_mode is not Strictness.HIGH:
                    builder.append(pack(possible[i]))
                i += 1

            return stroke + "".join(builder)

        return stroke + "".join([pack(tup) for tup in possible])


    def make_uncertainty_regex(self, stroke: str, uncertainty: int, annotations: list=list()) -> str:
        """Create a regular expression that matches a stroke within a given
        uncertainty while applying provided annotations.
        
        :param stroke: The stroke for which to generate alternatives.
        :param uncertainty: The uncertainty to apply to the stroke.
        :param annotations: A list of annotations to use in the generation.
        :returns: A regular expression.
        """

        similars = get_similar(stroke, uncertainty)
        flattened = []
        for group in similars:
            for token in group:
                flattened.append(self.make_annotation_regex(token, annotations))

        return "(" + "|".join(flattened) + ")"

    def build_regex(self, interpretation: Interpretation) -> str:
        """Create a regular expression from a grascii interpretation based
        on the constructor search parameters.
        
        :param interpretation: The interpretation for which to generate a 
            regular expression.
        :returns: A regular expression.
        """

        aspirate = "(" + re.escape(grammar.ASPIRATE) + ")"
        disjoiner = "(" + re.escape(grammar.DISJOINER) + ")"

        builder = list()
        i = 0
        if self.search_mode is SearchMode.MATCH or \
                self.search_mode is SearchMode.START:
            builder.append("^")
            if self.aspirate_mode is Strictness.LOW:
                builder.append(aspirate)
                builder.append("?")
            if self.aspirate_mode is Strictness.MEDIUM:
                while i < len(interpretation) and interpretation[i] == grammar.ASPIRATE:
                    builder.append(aspirate)
                    i += 1
                if i < 2:
                    builder.append(aspirate)
                    builder.append("?")

        if self.search_mode is SearchMode.CONTAIN:
            builder.append(".*")

        found_first = False

        while i < len(interpretation):
            token = interpretation[i]

            if token == grammar.ASPIRATE:
                if self.aspirate_mode is Strictness.MEDIUM or \
                        self.aspirate_mode is Strictness.HIGH:
                    builder.append(aspirate)
                i += 1
                continue

            if token == grammar.DISJOINER:
                if self.disjoiner_mode is Strictness.MEDIUM or \
                        self.disjoiner_mode is Strictness.HIGH:
                    builder.append(disjoiner)
                i += 1
                continue

            uncertainty = self.uncertainty if found_first or not self.fix_first else 0
            insert_aspirate = False
            if token in grammar.STROKES:
                if builder[-1] != re.escape(grammar.ASPIRATE) and self.aspirate_mode is Strictness.MEDIUM and found_first:
                    insert_aspirate = True
                if self.disjoiner_mode is Strictness.LOW and found_first:
                    builder.append(disjoiner)
                    builder.append("?")
                if builder[-1] != re.escape(grammar.DISJOINER) and self.disjoiner_mode is Strictness.MEDIUM and found_first:
                    builder.append(disjoiner)
                    builder.append("?")
                if self.aspirate_mode is Strictness.LOW:
                    builder.append(aspirate)
                    builder.append("?")
                if insert_aspirate:
                    builder.append(aspirate)
                    builder.append("?")
                found_first = True

            if token in grammar.ANNOTATIONS:
                if i + 1 < len(interpretation):
                    if isinstance(interpretation[i + 1], list):
                        builder.append(self.make_uncertainty_regex(token, uncertainty, interpretation[i + 1]))
                        i += 2
                        continue

            builder.append(self.make_uncertainty_regex(token, uncertainty))
            i += 1

        if self.search_mode is SearchMode.CONTAIN:
            builder.append(".*")

        if self.search_mode is SearchMode.MATCH or \
                self.search_mode is SearchMode.CONTAIN:
            builder.append(r"(?:\Z|\s)")

        return "".join(builder)

    def get_starting_letters(self, interpretations: List[Interpretation]) -> Set[str]:
        """Get a set of starting letters based on the given interpretations
        factoring in uncertainty.
        
        :param interpretations: A list of interpretations to generate 
            starting letters for.
        :returns: A set of characters.
        """

        if self.search_mode is SearchMode.CONTAIN:
            return grammar.HARD_CHARACTERS

        letters = set()
        for interp in interpretations:
            for token in interp:
                if token[0] in grammar.HARD_CHARACTERS:
                    if self.fix_first:
                        strokes = get_similar(token, 0)
                    else:
                        strokes = get_similar(token, self.uncertainty)
                    flattened_strokes = list()
                    for tup in strokes:
                        flattened_strokes += [s for s in tup]
                    letters |= set(string[0] for string in flattened_strokes)
                    break

        return letters

    def generate_patterns(self, interpretations: List[Interpretation]) -> List[Pattern]:
        """Generates a set of compiled regular expressions from a list
        of interpretations.
        
        :param interpretations: A list of interpretations to generate 
            patterns for.
        :returns: A list of Patterns.
        """

        patterns = list()
        for interp in interpretations:
            regex = self.build_regex(interp)
            # print(regex)
            patterns.append(re.compile(regex))
        return patterns


    def generate_patterns_map(self, interpretations: List[Interpretation]) -> List[Tuple[Interpretation, Pattern]]:
        """Generates a set of compiled regular expressions from a list
        of interpretations.
        
        :param interpretations: A list of interpretations to generate 
            patterns for.
        :returns: A list of Interpretations with their corresponding 
            Patterns.
        """

        patterns = list()
        for interp in interpretations:
            regex = self.build_regex(interp)
            # print(regex)
            # patterns.append(re.compile(regex))
            patterns.append((interp, re.compile(regex)))
        return patterns



"""
Contains metrics for comparing search queries to regular expression matches.
"""

import re
import string
from typing import List, Union, Set, Match, Tuple, cast

from grascii import grammar
from grascii.similarities import get_similar
from grascii.types import Interpretation


def convert_interpretation(interp: Interpretation) -> List[Tuple[str, Set[str]]]:
    """Convert an interpretation into a form for comparison.
    
    :param interp: The interpretation to convert.
    :returns: A standard form.
    """

    new_interp: List[Tuple[str, Set[str]]] = list()
    i = 0
    while i < len(interp):
        item = interp[i]
        if i + 1 < len(interp):
            if isinstance(interp[i + 1], list):
                new_interp.append((item, set(interp[i + 1])))
                i += 2
                continue
        new_interp.append((item, set()))
        i += 1
    return new_interp


def convert_match(match: Match) -> List[Tuple[str, Set[str]]]:
    """Convert an match into a form for comparison.
    
    :param match: The interpretation to convert.
    :returns: A standard form.
    """

    new_seq: List[Tuple[str, Set[str]]] = list()
    regex = re.compile("(" + "|".join(grammar.STROKES) + ")?(.*)")
    for group in match.groups():
        if group is None:
            continue
        i = 0
        while i < len(group) and group[i] in string.ascii_uppercase:
            i += 1
        if i > 0:
            new_seq.append((group[:i], {c for c in group[i:]}))
        else:
            new_seq.append((group, set()))
    return new_seq

def distance(s1: str, s2: str) -> int:
    """An implementation of the W-.. algorithm to calculate Lev distance
    based on the 2-vector Wikipedia pseudocode
    """

    m = len(s1)
    n = len(s2)

    v0 = [i for i in range(n + 1)]
    v1 = [0 for i in range(n + 1)]

    for i in range(m):
        v1[0] = i + 1
        
        for j in range(n):
            del_cost = v0[j + 1] + 1
            ins_cost = v1[j] + 1
            if s1[i] == s2[j]:
                sub_cost = v0[j]
            else:
                sub_cost = v0[j] + 1

            v1[j + 1] = min(del_cost, ins_cost, sub_cost)

        tmp = v0
        v0 = v1
        v1 = tmp
        
    return v0[n]

BASE_COST = 4

COSTS = {
    "~": 2,
    "|": 2,
    ".": 1,
    ",": 1,
    "(": 1,
    ")": 1,
    "_": 1
}
ASP_COST = 1
DIS_COST = BASE_COST

def compute_ins_del_cost(tup: Tuple[str, Set[str]]) -> int:
    """Compute the insertion and deletion cost of a stroke and its 
    annotations.
    
    :param tup: A stroke and its annotations for which to calculate its cost.
    :returns: A cost of insertion/deletion.
    """

    cost = 0
    if tup[0] in grammar.STROKES:
        cost += BASE_COST
    elif tup[0] == grammar.ASPIRATE:
        cost += ASP_COST
    elif tup[0] == grammar.DISJOINER:
        cost += BASE_COST

    if cost == 0:
         print(tup)
    assert cost > 0
    cost += sum(COSTS.get(a, 0) for a in tup[1])
    return cost

def get_distance(s1: str, s2: str) -> int:
    """Calculate the distance between two strokes in a similarity graph.
    Restricted to 0-3.

    :param s1: A stroke
    :param s2: A second stroke to compute the distance between
    :returns: A distance.
    """

    def in_sim(dis):
        sim = get_similar(s1, dis)
        s = set()
        for t in sim:
            s |= {a for a in t}
        return s2 in s

    for i in range(3):
        if in_sim(i):
            return i
    return 3

def compute_sub_cost(t1: Tuple[str, Set[str]], t2: Tuple[str, Set[str]]) -> int:
    """Compute the cost of substituting an annotated stroke with another one.
    
    :param t1: The stroke to replace.
    :param t2: The new stroke to add.
    :returns: A cost of substitution.
    """

    cost = 0
    if t1[0] in grammar.STROKES and t2[0] in grammar.STROKES:
        cost += get_distance(t1[0], t2[0]) * BASE_COST
        diff = t1[1] ^ t2[1]
        cost += sum(COSTS.get(a, 0) for a in diff)
        return cost
    else:
        if t1[0] == t2[0]:
            return 0
        return max(compute_ins_del_cost(t1), compute_ins_del_cost(t2))

def match_distance(g1: List[Tuple[str, Set[str]]], g2: List[Tuple[str, Set[str]]]) -> int:
    """Compute a weighed Lev distance between two sequences of annotated
    grascii tokens.
    
    :param g1: A grascii sequence.
    :param g2: A second grascii sequence.
    :returns: The edit distance between g1 and g2.
    """

    m = len(g1)
    n = len(g2)

    v0 = [0]
    for i in range(n):
        v0.append(v0[-1] + compute_ins_del_cost(g2[i]))
    v1 = [0 for i in range(n + 1)]

    for i in range(m):
        v1[0] = v0[0] + compute_ins_del_cost(g1[i])
        
        for j in range(n):
            del_cost = v0[j + 1] + compute_ins_del_cost(g1[i])
            ins_cost = v1[j] + compute_ins_del_cost(g2[j])
            sub_cost = v0[j] + compute_sub_cost(g1[i], g2[j])
            v1[j + 1] = min(del_cost, ins_cost, sub_cost)

        tmp = v0
        v0 = v1
        v1 = tmp
        
    return v0[n]

def standard(interp: Interpretation, match: Match) -> int:
    """Compute the standard metric for a grascii search.
    
    :param interp: The interpretation to compare to the match
    :param match: The search match.
    :returns: An edit distance.
    """

    g1 = convert_interpretation(interp)
    g2 = convert_match(match)
    return match_distance(g1, g2)



"""Contains grammars used by grascii."""

import io
from pathlib import Path
from pkg_resources import resource_string

def get_grammar(name: str) -> str:
    """Get a grammar string.
    
    :param name: The name of the grammar resource.
    :returns: A grammar string.
    """

    return Path(__file__).with_name(name + ".lark")

    # return resource_string("grascii.grammars", name + ".lark").decode("utf-8")



"""
Contains the base class for Searchers as well as multiple concrete 
implementations of it. 
"""

from abc import ABC, abstractmethod
from functools import reduce
import re
import sys
from typing import Union, List, Set, Iterable, Dict, TypeVar, Callable, Pattern, Tuple, Match

from lark import Lark, Tree, UnexpectedInput, Transformer, Token
from lark.visitors import CollapseAmbiguities

from grascii import regen, metrics, grammar, defaults
from grascii.dictionary import get_dict_file
from grascii.grammars import get_grammar
from grascii.types import Interpretation


class GrasciiFlattener(Transformer):

    """This is a Lark Transformer that converts a parsed Grascii string
    into an Interpretation. An Interpretation is a list of terminals and
    annotation lists. Each terminal is its own element in the interpretation,
    but sequences of annotation terminals are grouped into their own sublist.
    """

    def start(self, children) -> Interpretation:
        """Returns the final result of the transformation.
        
        :meta private:
        """

        result: Interpretation = list()
        for child in children:
            for token in child:
                if token in grammar.ANNOTATION_CHARACTERS:
                    # pack annotation terminals into sublist
                    if isinstance(result[-1], list):
                        # add to previous annotation list
                        result[-1].append(token)
                    else:
                        # start new annotation list
                        result.append([token])
                else:
                    result.append(token)
        return result

    def __default__(self, data, children, meta) -> List[str]:
        """The default visitor function for nodes in the parse tree.
        Returns a flat list of strings."""

        result: List[str] = list()
        for child in children:
            if isinstance(child, Token):
                result.append(child)
            else:
                # flatten iterable
                for token in child:
                    result.append(token)
        return result

T = TypeVar("T")

class Searcher(ABC):

    """An abstract base class for objects that search as Grascii dictionary."""

    def __init__(self, **kwargs):
        self.dictionary = kwargs.get("dictionary", defaults.SEARCH["Dictionary"])

    def perform_search(self, patterns: Iterable[Tuple[T, Pattern]], starting_letters: Set[str], metric: Callable[[T, Match], int]) -> Iterable[str]:
        """Performs a search of a Grascii Dictionary.
        
        :param patterns: A collection of compiled regular expression patterns
            with a corresponding interpretation.
        :param starting_letters: A set of letters used to index the search in
            a Grascii Dictionary.
        :param metric: A function taking an interpretation and a regular 
            expression match that returns a positive integer signifying the
            differnce between the interpretation and the match. 0 means the
            two are equivalent. The greater the value, the more different
            they are.
        :returns: A collection strings of the form "[grascii] [translation]"
            sorted by the results of metric.
        """
        sorted_results: List[Tuple[str, int]] = list()
        for item in sorted(starting_letters):
            try:
                with get_dict_file(self.dictionary, item) as dictionary:
                    for line in dictionary:
                        found_match = False
                        diff = 2^32 - 1
                        for interp, pattern in patterns:
                            match = pattern.search(line)
                            if match:
                                found_match = True
                                diff = min(diff, metric(interp, match))
                        if found_match:
                            i = len(sorted_results)
                            sorted_results.append((line, diff))
                            while i > 0:
                                if sorted_results[i][1] < sorted_results[i - 1][1]:
                                    tmp = sorted_results[i - 1]
                                    sorted_results[i - 1] = sorted_results[i]
                                    sorted_results[i] = tmp
                                i -= 1
            except FileNotFoundError:
                pass
                # print("Error: Could not find", dict_path + item)
        for tup in sorted_results:
            yield tup[0]

    @abstractmethod
    def search(self, **kwargs):
        """An abstract method that runs a search with the given search
        options and returns the results."""
        ...

class GrasciiSearcher(Searcher):

    """A subclass of Searcher that performs a search given a Grascii string"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        grammar = get_grammar("grascii")
        self.parser = Lark.open(grammar, parser="earley", ambiguity="explicit")

    def parse_grascii(self, grascii: str) -> Union[Tree, bool]:
        """Attempt to parse a grascii string.

        :param grascii: The grascii string to parse.
        :returns: A Lark parse tree on a successful parse, or False if the
            parse fails.
        """

        try:
            return self.parser.parse(grascii)
        except UnexpectedInput as e:
            print("Invalid Grascii String", file=sys.stderr)
            print(e.get_context(grascii), file=sys.stderr)
            return False

    def flatten_tree(self, parse_tree: Tree) -> List[Interpretation]:
        """Extract a list of all possible Interpretations from a grascii
            parse tree.

        :param parse_tree: A Lark parse tree from a parsed grascii string.
        :returns: A list of all possible Interpretations.
        """

        trees = CollapseAmbiguities().transform(parse_tree)
        trans = GrasciiFlattener()
        return [trans.transform(tree) for tree in trees]

    def interpretation_to_string(self, interpretation: Interpretation) -> str:
        """Generate a string representation of an Interpretation.
        
        :param interpretation: An Interpretation to generate a string for.
        :returns: A string representation of an Interpretation
        """

        def reducer(builder, token):
            if isinstance(token, list):
                builder += token
            else:
                if builder and builder[-1] != "^" and token != "^":
                    builder.append("-")
                builder.append(token)
            return builder

        return "".join(reduce(reducer, interpretation, []))

    def get_unique_interpretations(self, interpretations: List[Interpretation]) -> Dict[str, Interpretation]:
        """Generate a collection of all unique Interpretations in another
        collection by comparing string representations.
        
        :param interpretations: A collection of Interpretations to process
        :returns: A dictionary mapping interpretation string representations
            to Interpretations.
        """

        return {self.interpretation_to_string(interp): interp for interp in interpretations}

    def extract_search_args(self, **kwargs):
        """Get the relevant arguments for search."""

        self.uncertainty = kwargs.get("uncertainty", defaults.SEARCH.getint("Uncertainty"))
        # handle enum conversion error?
        try:
            self.search_mode = regen.SearchMode(kwargs.get("search_mode", defaults.SEARCH["SearchMode"]))
        except ValueError:
            self.search_mode = regen.SearchMode(defaults.DEFAULTS["Search"]["SearchMode"])
        try:
            self.annotation_mode = regen.Strictness(kwargs.get("annotation_mode", defaults.SEARCH["AnnotationMode"]))
        except ValueError:
            self.annotation_mode = regen.Strictness(defaults.DEFAULTS["Search"]["AnnotationMode"])
        try:
            self.aspirate_mode = regen.Strictness(kwargs.get("aspirate_mode", defaults.SEARCH["AspirateMode"]))
        except ValueError:
            self.aspirate_mode = regen.Strictness(defaults.DEFAULTS["Search"]["AspirateMode"])
        try:
            self.disjoiner_mode = regen.Strictness(kwargs.get("disjoiner_mode", defaults.SEARCH["DisjoinerMode"]))
        except:
            self.disjoiner_mode = regen.Strictness(defaults.DEFAULTS["Search"]["DisjoinerMode"])
        self.fix_first = kwargs.get("fix_first", False)
        self.interpretation_mode = kwargs.get("interpretation", defaults.SEARCH["Interpretation"])

    def search(self, **kwargs):
        """
        :param grascii: [Required] The grascii string to use in the search.
        :param uncertainty: The uncertainty of the grascii string.
        :param search_mode: The search mode to use.
        :param annotation_mode: How to handle annotations in the search.
        :param aspirate_mode: How to handle annotations in the search.
        :param disjoiner_mode: How to handle annotations in the search.
        :param fix_first: Apply an uncertainty of 0 to the first token.
        :param interpretation: How to handle ambiguous grascii strings.
        :type grascii: str
        :type uncertainty: int: 0, 1, or 2
        :type search_mode: str: one of regen.SearchMode values
        :type annotation_mode: one of regen.Strictness values
        :type aspirate_mode: one of regen.Strictness values
        :type disjoiner_mode: one of regen.Strictness values
        :type fix_first: bool
        :type interpretation: "best" or "all"
        :returns: A list of search results.
        :rtype: List[str]
        """

        grascii = kwargs["grascii"].upper()
        self.extract_search_args(**kwargs)
        tree = self.parse_grascii(grascii)
        if not tree:
            return

        interpretations = self.flatten_tree(tree)
        interpretations = list(self.get_unique_interpretations(interpretations).values())
        builder = regen.RegexBuilder(
            uncertainty=self.uncertainty,
            search_mode=self.search_mode,
            aspirate_mode=self.aspirate_mode,
            annotation_mode=self.annotation_mode,
            disjoiner_mode=self.disjoiner_mode,
            fix_first=self.fix_first
        )

        interps = interpretations[0:1] if self.interpretation_mode == "best" else interpretations
        patterns = builder.generate_patterns_map(interps)
        starting_letters = builder.get_starting_letters(interps)

        results = self.perform_search(patterns, starting_letters, metrics.standard)
        return list(results)

class RegexSearcher(Searcher):

    """A subclass of Searcher that searches a grascii dictionary given
    a raw regular expression."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def search(self, **kwargs):
        """
        :param regexp: [Required] A regular expression to use in a search.
        :returns: A list of search results.
        :rtype: List[str]
        """

        regex = kwargs["regexp"]
        print(regex)
        pattern = re.compile(regex)
        patterns = [(pattern.pattern, pattern)]
        metric = lambda str, Match: 0
        starting_letters = grammar.HARD_CHARACTERS
        results = self.perform_search(patterns, starting_letters, metric)
        return list(results)

class ReverseSearcher(RegexSearcher):

    """A subclass of RegexSearcher that searches a grascii dictionary
    given a word."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def search(self, **kwargs):
        """
        :param reverse: [Required] A word to search for.
        :returns: A list of search results.
        :rtype: List[str]
        """

        word = kwargs["reverse"]
        kwargs["regexp"] = r".*\s" + word.capitalize()
        return super().search(**kwargs)




import re
import unittest

from typing import List, Tuple

from grascii import regen, grammar, similarities

class TestAnnotationRegex(unittest.TestCase):

    def check_strictness_low(self, annotations: List[str], texts: List[Tuple[str, str]]):
        builder = regen.RegexBuilder(annotation_mode=regen.Strictness.LOW)
        for a in annotations:
            for t in texts:
                regex = builder.make_annotation_regex(t[0], a)
                self.assertRegex(t[0] + t[1], regex)

    def test_strictness_low_circle_vowel(self):
        annotations = [
            [],
            [","],
            ["~", ","]
        ]
        texts = [("A", ""), ("A", "~"), ("A", "|."), ("A", "~|,_"),
                 ("E", ""), ("E", "~"), ("E", "|."), ("E", "~|,_"),]
        self.check_strictness_low(annotations, texts)

    def test_strictness_low_hook_vowel(self):
        annotations_o = [
            [],
            [","],
            ["(", ","],
            ["(", "_"]
        ]
        texts_o = [("O", ""), ("O", "("), ("O", "(,"), ("O", "(._")]
        annotations_u = [
            [],
            [","],
            [")", ","],
            [")", "_"]
        ]
        texts_u = [("U", ""), ("U", ")"), ("U", "),"), ("U", ")._")]
        self.check_strictness_low(annotations_o, texts_o)
        self.check_strictness_low(annotations_u, texts_u)

    def test_strictness_low_circle_diphthong(self):
        annotations = [
            [],
            ["~"],
            ["~", "|"],
            ["_"],
        ]
        texts = [("I", ""), ("I", "~"), ("I", "_"), ("I~|_"),
                ("A&E", ""), ("A&E", "~"), ("A&E", "_"), ("A&E~|_"),    
                ("A'E", ""), ("A'E", "~"), ("A'E", "_"), ("A'E~|_")]
        self.check_strictness_low(annotations, texts)
                
    def test_strictness_low_hook_diphthong(self):
        annotations = [
            [],
            ["_"],
        ]
        texts = [("AU", ""), ("AU", "_"),
                 ("OE", ""), ("OE", "_"),
                 ("EU", ""), ("EU", "_")]
        self.check_strictness_low(annotations, texts)

    def test_strictness_low_directed_consonant(self):
        annotations = [
            [],
            ["("],
            [")"],
            [","],
            ["(,"]
        ]
        texts = [("S", ""), ("S", "("), ("S", "),"),
                ("Z", ""), ("Z", "("), ("Z", "),"),
                ("TH", ""), ("TH", "("), ("TH", "),")]
        self.check_strictness_low(annotations, texts)

    def test_strictness_low_oblique_consonant(self):
        annotations = [
            [],
            [","]
        ]
        texts = [("SH", ""), ("SH", ",")]
        self.check_strictness_low(annotations, texts)

    def check_strictness_medium(self, strokes, tests):
        builder = regen.RegexBuilder(annotation_mode=regen.Strictness.MEDIUM)
        for stroke in strokes:
            for test in tests:
                annotations = test[0]
                for text, expected in test[1]:
                    regex = builder.make_annotation_regex(stroke, annotations)
                    with self.subTest(annotations=annotations, text=stroke+text):
                        if expected:
                            self.assertRegex(stroke + text, regex)
                        else:
                            self.assertNotRegex(stroke + text, regex)


    def test_strictness_medium_circle_vowel(self):
        circle_vowels = ["A", "E"]
        tests = [
            ([], [("", True), (",", True), ("~|._", True)]),
            ([","], [("", False), (".", False), (",", True), ("~|,_", True)]),
            (["~"], [("", False), ("~", True), (".", False), ("~|._", True)]),
            (["~", "|", ".", "_"], [("~", False), ("|", False), (".", False),
                ("_", False), ("~|._", True)]),
        ]
        self.check_strictness_medium(circle_vowels, tests)

    def test_strictness_medium_hook_vowel(self):
        tests1 = [
            ([], [("", True), (",", True), ("(,_", True)]),
            ([","], [("", False), (".", False), (",", True), ("(,_", True)]),
            (["("], [("", False), (".", False), ("(,_", True)]),
            (["(", ".", "_"], [(".", False), ("_", False), ("(._", True)]),
        ]
        tests2 = [
            ([], [("", True), (",", True), ("),_", True)]),
            ([","], [("", False), (".", False), (",", True), ("),_", True)]),
            ([")"], [("", False), (".", False), ("),_", True)]),
            ([")", ".", "_"], [(".", False), ("_", False), (")._", True)]),
        ]
        self.check_strictness_medium(["O"], tests1)
        self.check_strictness_medium(["U"], tests2)

    def test_strictness_medium_circle_diphthong(self):
        strokes = ["I", "A&E", "A&'"]
        tests = [
            ([], [("", True), ("~", True), ("|", True), ("_", True), ("~|_", True)]),
            (["~"], [("", False), ("~", True), ("|", False), ("_", False), ("~|_", True)]),
            (["|"], [("", False), ("~", False), ("|", True), ("_", False), ("~|_", True)]),
            (["_"], [("", False), ("~", False), ("|", False), ("_", True), ("~|_", True)]),
            (["~", "|", "_"], [("", False), ("~", False), ("|", False), ("_", False), ("~|_", True)]),
        ]
        self.check_strictness_medium(strokes, tests)

    def test_strictness_medium_hook_diphthong(self):
        strokes = ["AU", "OE", "EU"]
        tests = [
            ([], [("", True), ("_", True)]),
            (["_"], [("", False), ("_", True)]),
        ]
        self.check_strictness_medium(strokes, tests)

    def test_strictness_medium_directed_consonant(self):
        strokes = ["S", "Z", "TH"]
        tests = [
            ([], [("", True), ("(", True), (")", True), (",", True), ("(,", True), ("),", True)]),
            (["("], [("", False), ("(", True), (")", False), (",", False), ("(,", True), ("),", False)]),
            ([")"], [("", False), ("(", False), (")", True), (",", False), ("(,", False), ("),", True)]),
            (["(", ","], [("", False), ("(", False), (")", False), (",", False), ("(,", True), ("),", False)]),
            ([")", ","], [("", False), ("(", False), (")", False), (",", False), ("(,", False), ("),", True)])
        ]
        self.check_strictness_medium(strokes, tests)

    def test_strictness_medium_oblique_consonant(self):
        tests = [
            ([], [("", True), (",", True)]),
            ([","], [("", False), (",", True)]),
        ]
        self.check_strictness_medium(["SH"], tests)

    def check_strictness_high(self, strokes, tests):
        builder = regen.RegexBuilder(annotation_mode=regen.Strictness.HIGH)
        for stroke in strokes:
            for test in tests:
                annotations = test[0]
                for text, expected in test[1]:
                    regex = builder.make_annotation_regex(stroke, annotations)
                    with self.subTest(annotations=annotations, text=stroke+text):
                        if expected:
                            self.assertTrue(re.fullmatch(regex, stroke + text))
                        else:
                            self.assertFalse(re.fullmatch(regex, stroke + text))


    def test_strictness_high_circle_vowel(self):
        circle_vowels = ["A", "E"]
        tests = [
            ([], [("", True), (",", False), ("~|._", False)]),
            ([","], [("", False), (".", False), (",", True), ("~|,_", False)]),
            (["~"], [("", False), ("~", True), (".", False), ("~|._", False)]),
            (["~", "|", ".", "_"], [("~", False), ("|", False), (".", False),
                ("_", False), ("~|._", True)]),
        ]
        self.check_strictness_high(circle_vowels, tests)

    def test_strictness_high_hook_vowel(self):
        tests1 = [
            ([], [("", True), (",", False), ("(,_", False)]),
            ([","], [("", False), (".", False), (",", True), ("(,_", False)]),
            (["("], [("", False), (".", False), ("(,_", False)]),
            (["(", ".", "_"], [(".", False), ("_", False), ("(._", True)]),
        ]
        tests2 = [
            ([], [("", True), (",", False), ("),_", False)]),
            ([","], [("", False), (".", False), (",", True), ("),_", False)]),
            ([")"], [("", False), (".", False), ("),_", False)]),
            ([")", ".", "_"], [(".", False), ("_", False), (")._", True)]),
        ]
        self.check_strictness_high(["O"], tests1)
        self.check_strictness_high(["U"], tests2)

    def test_strictness_high_circle_diphthong(self):
        strokes = ["I", "A&E", "A&'"]
        tests = [
            ([], [("", True), ("~", False), ("|", False), ("_", False), ("~|_", False)]),
            (["~"], [("", False), ("~", True), ("|", False), ("_", False), ("~|_", False)]),
            (["|"], [("", False), ("~", False), ("|", True), ("_", False), ("~|_", False)]),
            (["_"], [("", False), ("~", False), ("|", False), ("_", True), ("~|_", False)]),
            (["~", "|", "_"], [("", False), ("~", False), ("|", False), ("_", False), ("~|_", True)]),
        ]
        self.check_strictness_high(strokes, tests)

    def test_strictness_high_hook_diphthong(self):
        strokes = ["AU", "OE", "EU"]
        tests = [
            ([], [("", True), ("_", False)]),
            (["_"], [("", False), ("_", True)]),
        ]
        self.check_strictness_high(strokes, tests)

    def test_strictness_high_directed_consonant(self):
        strokes = ["S", "Z", "TH"]
        tests = [
            ([], [("", True), ("(", False), (")", False), (",", False), ("(,", False), ("),", False)]),
            (["("], [("", False), ("(", True), (")", False), (",", False), ("(,", False), ("),", False)]),
            ([")"], [("", False), ("(", False), (")", True), (",", False), ("(,", False), ("),", False)]),
            (["(", ","], [("", False), ("(", False), (")", False), (",", False), ("(,", True), ("),", False)]),
            ([")", ","], [("", False), ("(", False), (")", False), (",", False), ("(,", False), ("),", True)])
        ]
        self.check_strictness_high(strokes, tests)

    def test_strictness_high_oblique_consonant(self):
        tests = [
            ([], [("", True), (",", False)]),
            ([","], [("", False), (",", True)]),
        ]
        self.check_strictness_high(["SH"], tests)

    def test_out_of_order_annotations(self):
        pass

    def test_invalid_annotations(self):
        pass

    def test_mutually_exclusive_annotations(self):
        pass



class TestDisjoiners(unittest.TestCase):

    def run_tests(self, builder, tests):
        for test in tests:
            interp = test[0]
            for text, expected in test[1]:
                regex = builder.build_regex(interp)
                with self.subTest(interpretation=interp, text=text):
                    if expected:
                        self.assertRegex(text, regex)
                    else:
                        self.assertNotRegex(text, regex)

    def test_strictness_low(self):
        builder = regen.RegexBuilder(disjoiner_mode=regen.Strictness.LOW)
        tests = [
            (["A", "B"], [("AB", True), ("A^B", True)]),
            (["A", "^", "B"], [("AB", True), ("A^B", True)]),
            (["A", "B", "D"], [("ABD", True), ("A^BD", True), ("AB^D", True), ("A^B^D", True)]),
            (["A", "^", "B", "D"], [("ABD", True), ("A^BD", True), ("AB^D", True), ("A^B^D", True)]),
            (["A", "B", "^", "D"], [("ABD", True), ("A^BD", True), ("AB^D", True), ("A^B^D", True)]),
            (["A", "^", "B", "^", "D"], [("ABD", True), ("A^BD", True), ("AB^D", True), ("A^B^D", True)])
        ]
        self.run_tests(builder, tests)

    def test_strictness_medium(self):
        builder = regen.RegexBuilder(disjoiner_mode=regen.Strictness.MEDIUM)
        tests = [
            (["A", "B"], [("AB", True), ("A^B", True)]),
            (["A", "^", "B"], [("AB", False), ("A^B", True)]),
            (["A", "B", "D"], [("ABD", True), ("A^BD", True), ("AB^D", True), ("A^B^D", True)]),
            (["A", "^", "B", "D"], [("ABD", False), ("A^BD", True), ("AB^D", False), ("A^B^D", True)]),
            (["A", "B", "^", "D"], [("ABD", False), ("A^BD", False), ("AB^D", True), ("A^B^D", True)]),
            (["A", "^", "B", "^", "D"], [("ABD", False), ("A^BD", False), ("AB^D", False), ("A^B^D", True)])
        ]
        self.run_tests(builder, tests)

    def test_strictness_high(self):
        builder = regen.RegexBuilder(disjoiner_mode=regen.Strictness.HIGH)
        tests = [
            (["A", "B"], [("AB", True), ("A^B", False)]),
            (["A", "^", "B"], [("AB", False), ("A^B", True)]),
            (["A", "B", "D"], [("ABD", True), ("A^BD", False), ("AB^D", False), ("A^B^D", False)]),
            (["A", "^", "B", "D"], [("ABD", False), ("A^BD", True), ("AB^D", False), ("A^B^D", False)]),
            (["A", "B", "^", "D"], [("ABD", False), ("A^BD", False), ("AB^D", True), ("A^B^D", False)]),
            (["A", "^", "B", "^", "D"], [("ABD", False), ("A^BD", False), ("AB^D", False), ("A^B^D", True)])
        ]
        self.run_tests(builder, tests)

    def test_prefixes(self):
        pass

class TestAspirates(unittest.TestCase):

    def run_tests(self, builder, tests):
        for test in tests:
            interp = test[0]
            for text, expected in test[1]:
                regex = builder.build_regex(interp)
                with self.subTest(interpretation=interp, text=text):
                    if expected:
                        self.assertRegex(text, regex)
                    else:
                        self.assertNotRegex(text, regex)

    def test_strictness_low(self):
        builder = regen.RegexBuilder(aspirate_mode=regen.Strictness.LOW)
        tests = [
            (["A"], [("A", True), ("'A", True)]),
            (["'", "A"], [("A", True), ("'A", True)]),
            (["A", "D", "E"], [("ADE", True), ("'ADE", True), ("A'DE", True), ("AD'E", True), ("'A'DE", True),
                ("'AD'E", True), ("A'D'E", True), ("'A'D'E", True)]),
            (["'", "A", "D", "E"], [("ADE", True), ("'ADE", True), ("A'DE", True), ("AD'E", True), ("'A'DE", True),
                ("'AD'E", True), ("A'D'E", True), ("'A'D'E", True)]),
            (["'", "A", "'", "D", "E"], [("ADE", True), ("'ADE", True), ("A'DE", True), ("AD'E", True), ("'A'DE", True),
                ("'AD'E", True), ("A'D'E", True), ("'A'D'E", True)]),
            (["'", "A", "'", "D", "'", "E"], [("ADE", True), ("'ADE", True), ("A'DE", True), ("AD'E", True), ("'A'DE", True),
                ("'AD'E", True), ("A'D'E", True), ("'A'D'E", True)]),
            (["A", "'", "D", "'", "E"], [("ADE", True), ("'ADE", True), ("A'DE", True), ("AD'E", True), ("'A'DE", True),
                ("'AD'E", True), ("A'D'E", True), ("'A'D'E", True)]),
            (["'", "A", "D", "'", "E"], [("ADE", True), ("'ADE", True), ("A'DE", True), ("AD'E", True), ("'A'DE", True),
                ("'AD'E", True), ("A'D'E", True), ("'A'D'E", True)]),
            (["A", "'", "D", "E"], [("ADE", True), ("'ADE", True), ("A'DE", True), ("AD'E", True), ("'A'DE", True),
                ("'AD'E", True), ("A'D'E", True), ("'A'D'E", True)]),
            (["A", "D", "'", "E"], [("ADE", True), ("'ADE", True), ("A'DE", True), ("AD'E", True), ("'A'DE", True),
                ("'AD'E", True), ("A'D'E", True), ("'A'D'E", True)]),
        ]
        self.run_tests(builder, tests)

    def test_strictness_medium(self):
        builder = regen.RegexBuilder(aspirate_mode=regen.Strictness.MEDIUM)
        tests = [
            (["A"], [("A", True), ("'A", True)]),
            (["'", "A"], [("A", False), ("'A", True)]),
            (["A", "D", "E"], [("ADE", True), ("'ADE", True), ("A'DE", True), ("AD'E", True), ("'A'DE", True),
                ("'AD'E", True), ("A'D'E", True), ("'A'D'E", True)]),
            (["'", "A", "D", "E"], [("ADE", False), ("'ADE", True), ("A'DE", False), ("AD'E", False), ("'A'DE", True),
                ("'AD'E", True), ("A'D'E", False), ("'A'D'E", True)]),
            (["'", "A", "'", "D", "E"], [("ADE", False), ("'ADE", False), ("A'DE", False), ("AD'E", False), ("'A'DE", True),
                ("'AD'E", False), ("A'D'E", False), ("'A'D'E", True)]),
            (["'", "A", "'", "D", "'", "E"], [("ADE", False), ("'ADE", False), ("A'DE", False), ("AD'E", False), ("'A'DE", False),
                ("'AD'E", False), ("A'D'E", False), ("'A'D'E", True)]),
            (["A", "'", "D", "'", "E"], [("ADE", False), ("'ADE", False), ("A'DE", False), ("AD'E", False), ("'A'DE", False),
                ("'AD'E", False), ("A'D'E", True), ("'A'D'E", True)]),
            (["'", "A", "D", "'", "E"], [("ADE", False), ("'ADE", False), ("A'DE", False), ("AD'E", False), ("'A'DE", False),
                ("'AD'E", True), ("A'D'E", False), ("'A'D'E", True)]),
            (["A", "'", "D", "E"], [("ADE", False), ("'ADE", False), ("A'DE", True), ("AD'E", False), ("'A'DE", True),
                ("'AD'E", False), ("A'D'E", True), ("'A'D'E", True)]),
            (["A", "D", "'", "E"], [("ADE", False), ("'ADE", False), ("A'DE", False), ("AD'E", True), ("'A'DE", False),
                ("'AD'E", True), ("A'D'E", True), ("'A'D'E", True)]),
        ]
        self.run_tests(builder, tests)

    def test_strictness_high(self):
        builder = regen.RegexBuilder(aspirate_mode=regen.Strictness.HIGH)
        tests = [
            (["A"], [("A", True), ("'A", False)]),
            (["'", "A"], [("A", False), ("'A", True)]),
            (["A", "D", "E"], [("ADE", True), ("'ADE", False), ("A'DE", False), ("AD'E", False), ("'A'DE", False),
                ("'AD'E", False), ("A'D'E", False), ("'A'D'E", False)]),
            (["'", "A", "D", "E"], [("ADE", False), ("'ADE", True), ("A'DE", False), ("AD'E", False), ("'A'DE", False),
                ("'AD'E", False), ("A'D'E", False), ("'A'D'E", False)]),
            (["'", "A", "'", "D", "E"], [("ADE", False), ("'ADE", False), ("A'DE", False), ("AD'E", False), ("'A'DE", True),
                ("'AD'E", False), ("A'D'E", False), ("'A'D'E", False)]),
            (["'", "A", "'", "D", "'", "E"], [("ADE", False), ("'ADE", False), ("A'DE", False), ("AD'E", False), ("'A'DE", False),
                ("'AD'E", False), ("A'D'E", False), ("'A'D'E", True)]),
            (["A", "'", "D", "'", "E"], [("ADE", False), ("'ADE", False), ("A'DE", False), ("AD'E", False), ("'A'DE", False),
                ("'AD'E", False), ("A'D'E", True), ("'A'D'E", False)]),
            (["'", "A", "D", "'", "E"], [("ADE", False), ("'ADE", False), ("A'DE", False), ("AD'E", False), ("'A'DE", False),
                ("'AD'E", True), ("A'D'E", False), ("'A'D'E", False)]),
            (["A", "'", "D", "E"], [("ADE", False), ("'ADE", False), ("A'DE", True), ("AD'E", False), ("'A'DE", False),
                ("'AD'E", False), ("A'D'E", False), ("'A'D'E", False)]),
            (["A", "D", "'", "E"], [("ADE", False), ("'ADE", False), ("A'DE", False), ("AD'E", True), ("'A'DE", False),
                ("'AD'E", False), ("A'D'E", False), ("'A'D'E", False)]),
        ]
        self.run_tests(builder, tests)

    def test_double_aspirate(self):
        pass
        
class TestUncertaintyRegex(unittest.TestCase):
    
    def run_tests(self, uncertainty, tests):
        builder = regen.RegexBuilder()
        for test in tests:
            stroke = test[0]
            for text, expected in test[1]:
                regex = builder.make_uncertainty_regex(stroke, uncertainty)
                with self.subTest():
                    if expected:
                        self.assertTrue(re.fullmatch(regex, text))
                    else:
                        self.assertFalse(re.fullmatch(regex, text))

    # def test_uncertainty_zero(self):
        # # this should go under sim tests
        # for stroke in grammar.STROKES:
            # similars = similarities.get_similar(stroke, 0)
            # self.assertEqual(len(similars), 1)
            # self.assertIn(stroke, list(similars)[0])

    def test_uncertainty_zero(self):
        tests = [
            ("A", [("A", True), ("E", False), ("I", False)]),
            ("NT", [("NT", True), ("ND", True), ("TH", False), ("MT", False), ("MD", False)]),
            ("F", [("F", True), ("V", False), ("CH", False), ("S", False),
                   ("Z", False), ("SH", False), ("P", False), ("J", False),
                   ("X", False), ("XS", False), ("SS", False), ("B", False)])
        ]
        self.run_tests(0, tests)

    def test_uncertainty_one(self):
        tests = [
            ("A", [("A", True), ("E", True), ("I", True)]),
            ("NT", [("NT", True), ("ND", True), ("TH", True), ("MT", True), ("MD", True)]),
            ("F", [("F", True), ("V", True), ("CH", True), ("S", True),
                   ("Z", True), ("SH", False), ("P", False), ("J", False),
                   ("X", False), ("XS", False), ("SS", False), ("B", False)])
        ]
        self.run_tests(1, tests)

    def test_uncertainty_two(self):
        tests = [
            ("A", [("A", True), ("E", True), ("I", True)]),
            ("NT", [("NT", True), ("ND", True), ("TH", True), ("MT", True), ("MD", True)]),
            ("F", [("F", True), ("V", True), ("CH", True), ("S", True),
                   ("Z", True), ("SH", True), ("P", True), ("J", True),
                   ("X", True), ("XS", True), ("SS", True), ("B", False)])
        ]
        self.run_tests(2, tests)


class TestSearchMode(unittest.TestCase):

    def run_tests(self, builder, tests):
        for test in tests:
            interp = test[0]
            for text, expected in test[1]:
                regex = builder.build_regex(interp)
                with self.subTest(interpretation=interp, text=text):
                    if expected:
                        self.assertRegex(text, regex)
                    else:
                        self.assertNotRegex(text, regex)
    
    def test_match(self):
        builder = regen.RegexBuilder(search_mode=regen.SearchMode.MATCH)
        tests = [
            (["A", "B"], [("AB", True), ("ABU", False), ("DAB", False)])
        ]
        self.run_tests(builder, tests)

    def test_start(self):
        builder = regen.RegexBuilder(search_mode=regen.SearchMode.START)
        tests = [
            (["A", "B"], [("AB", True), ("ABU", True), ("DAB", False)])
        ]
        self.run_tests(builder, tests)

    def test_contains(self):
        builder = regen.RegexBuilder(search_mode=regen.SearchMode.CONTAIN)
        tests = [
            (["A", "B"], [("AB", True), ("ABU", True), ("DAB", True)])
        ]
        self.run_tests(builder, tests)

class TestFixFirst(unittest.TestCase):

    def run_tests(self, builder, tests):
        for test in tests:
            interp = test[0]
            for text, expected in test[1]:
                regex = builder.build_regex(interp)
                with self.subTest(interpretation=interp, text=text):
                    if expected:
                        self.assertRegex(text, regex)
                    else:
                        self.assertNotRegex(text, regex)

    def test_fix_first_off(self):
        builder = regen.RegexBuilder(fix_first=False, uncertainty=1)
        tests = [
            (["A", "B", "D"], [("ABD", True), ("EBD", True), ("IBD", True),
                               ("APT", True), ("EPDD", True), ("IBDT", True)])
        ]
        self.run_tests(builder, tests)

    def test_fix_first_on(self):
        builder = regen.RegexBuilder(fix_first=True, uncertainty=1)
        tests = [
            (["A", "B", "D"], [("ABD", True), ("EBD", False), ("IBD", False),
                               ("APT", True), ("EPDD", False), ("IBDT", False)])
        ]
        self.run_tests(builder, tests)

class TestStartingLetters(unittest.TestCase):
    
    def run_tests(self, builder, tests):
        for test in tests:
            interps = test[0]
            for stroke, expected in test[1]:
                letters = builder.get_starting_letters(interps)
                with self.subTest(interpretations=interps, stroke=stroke):
                    if expected:
                        self.assertIn(stroke, letters)
                    else:
                        self.assertNotIn(stroke, letters)

    def test_starting_letters_search_mode_contains(self):
        builder = regen.RegexBuilder(search_mode=regen.SearchMode.CONTAIN)
        tests = [[["A"]], [["B"]], [["K", "P"]], [["'", "I"]]]
        for test in tests:
            letters = builder.get_starting_letters(test)
            self.assertSetEqual(letters, grammar.HARD_CHARACTERS)

    def test_starting_letters_fix_first_uncertainty_one(self):
        builder = regen.RegexBuilder(search_mode=regen.SearchMode.MATCH,
                                     uncertainty=1,
                                     fix_first=True)
        tests = [
            ([["A"]], [("A", True), ("E", False), ("I", False)]),
            ([["'", "I"], [("I", True), ("E", False), ("A", False)]]),
            ([["NT"], ["DN"]], [("N", True), ("T", True), ("D", True),
                ("M", False)])
        ]
        self.run_tests(builder, tests)

    def test_starting_letters_no_fix_first_uncertainty_one(self):
        builder = regen.RegexBuilder(search_mode=regen.SearchMode.MATCH,
                                     uncertainty=1,
                                     fix_first=False)
        tests = [
            ([["A"]], [("A", True), ("E", True), ("I", True)]),
            ([["'", "I"], [("I", True), ("E", False), ("A", True)]]),
            ([["NT"], ["DN"]], [("N", True), ("T", True), ("D", True),
                ("M", True)])
        ]
        self.run_tests(builder, tests)


class TestRegexBuilder(unittest.TestCase):
    pass


         

    

if __name__ == "__main__":
    unittest.main()



"""
Acts as the main entry point for the grascii search command.

This can be invoked as a standalone program:
$ python -m grascii.search --help
"""

import argparse
from configparser import ConfigParser
import io
import os
import re
import sys
from typing import Union, List, Dict, Set, Iterable, Match, Pattern, cast, Optional

from grascii import regen, defaults
from grascii.searchers import GrasciiSearcher, RegexSearcher, ReverseSearcher, Searcher

SUPPORTS_INTERACTIVE = False
try:
    from grascii.interactive import InteractiveSearcher
    SUPPORTS_INTERACTIVE = True
except ImportError as e:
    pass


description = "Search a Grascii Dictionary"

def build_argparser(argparser: argparse.ArgumentParser) -> None:
    """Configure an ArgumentParser parser to parse the search command-line
    options.
    
    :param argparser: A fresh ArgumentParser to configure.
    """

    group = argparser.add_mutually_exclusive_group(required=True)
    group.add_argument("-g", "--grascii", 
            help="the grascii string to search for")
    group.add_argument("-e", "--regexp", 
            help="a custom regular expression to use in the search")
    group.add_argument("-r", "--reverse", 
            help="search by word instead")
    group.add_argument("-i", "--interactive", action="store_true",
            help="run in interactive mode")
    argparser.add_argument("-u", "--uncertainty", type=int, choices=range(3),
            help="the uncertainty of the search term")
    argparser.add_argument("-s", "--search-mode",
            choices=[mode.value for mode in regen.SearchMode],
            help="the kind of search to perform")
    argparser.add_argument("-a", "--annotation-mode",
            choices=[mode.value for mode in regen.Strictness],
            help="how to handle annotations in grascii")
    argparser.add_argument("-p", "--aspirate-mode",
            choices=[mode.value for mode in regen.Strictness],
            help="how to handle aspirates in grascii")
    argparser.add_argument("-j", "--disjoiner-mode",
            choices=[mode.value for mode in regen.Strictness],
            help="how to handle disjoiners in grascii")
    argparser.add_argument("-n", "--interpretation", 
            choices=["best", "all"],
            help="how to handle ambiguous grascii strings")
    argparser.add_argument("-f", "--fix-first", action="store_true",
            help="apply an uncertainty of 0 to the first token")
    argparser.add_argument("-d", "--dictionary", 
            help="set the dictionary to be searched")

def search(**kwargs) -> Optional[Iterable[str]]:
    """Run a grascii dictionary search. Parameters can consist of
    any parameters used by the search method of any subclass of
    Searcher. One, and only one, of the parameters list below
    is required.

    :param grascii: A grascii string to use in a search.
    :param interactive: A flag enabling an interactive search.
    :param reverse: A word to search for in the dictionary.
    :param regexp: A regular expression to use in a search.
    :type grascii: str
    :type interactive: bool
    :type reverse: str
    :type regexp: str

    :returns: A list of search results, or None if run in interactive mode
    """

    searcher: Searcher
    if kwargs.get("grascii"):
        searcher = GrasciiSearcher(**kwargs)
    elif kwargs.get("interactive"):
        if not SUPPORTS_INTERACTIVE:
            print("The interactive extra is not installed", file=sys.stderr)
            return
        searcher = InteractiveSearcher(**kwargs)
    elif kwargs.get("reverse"):
        searcher = ReverseSearcher(**kwargs)
    else:
        searcher = RegexSearcher(**kwargs)
    return searcher.search(**kwargs)

def cli_search(args: argparse.Namespace) -> None:
    """Run a search using arguments parsed from the command line.
    
    :param args: A namespace of parsed arguments.
    """

    results = search(**{k: v for k, v in vars(args).items() if v is not None})
    if results is not None:
        count = 0
        for result in results:
            print(result.strip())
            count += 1
        print("Results:", count)

def main() -> None:
    """Run a search using arguments retrieved from sys.argv."""

    argparser = argparse.ArgumentParser(description)
    build_argparser(argparser)
    args = argparser.parse_args(sys.argv[1:])
    cli_search(args)
    search(**{k: v for k, v in vars(args).items() if v is not None})

if __name__ == "__main__":
    main()



"""
Contains an implementation of an Interactive Searcher for an 
interactive cli experience.
"""

try:
    import questionary
    from questionary.prompts.common import Choice, Separator
except ImportError:
    raise ImportError("Grascii: interactive extra dependencies are not installed")

from lark import Tree
from typing import Iterable, Sequence, TypeVar, Callable, Optional, Tuple

from grascii import regen, metrics
from grascii.searchers import GrasciiSearcher, Interpretation

T = TypeVar("T")

class InteractiveSearcher(GrasciiSearcher):
    """This subclass of GrasciiSearcher runs an interactive search
    experience for performing grascii searches with support for changing
    search parameters.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def search(self, **kwargs):
        """
        :param grascii: [Required] The grascii string to use in the search.
        :param uncertainty: The uncertainty of the grascii string.
        :param search_mode: The search mode to use.
        :param annotation_mode: How to handle annotations in the search.
        :param aspirate_mode: How to handle annotations in the search.
        :param disjoiner_mode: How to handle annotations in the search.
        :param fix_first: Apply an uncertainty of 0 to the first token.
        :type grascii: str
        :type uncertainty: int: 0, 1, or 2
        :type search_mode: str: one of regen.SearchMode values
        :type annotation_mode: one of regen.Strictness values
        :type aspirate_mode: one of regen.Strictness values
        :type disjoiner_mode: one of regen.Strictness values
        :type fix_first: bool
        :returns: None
        :rtype: None
        """

        self.extract_search_args(**kwargs)
        self.run_interactive()

    def choose_interpretation(self, interpretations: Sequence[Interpretation]) -> int:
        """Prompt the user to choose an interpretation(s).
        
        :param interpretations: A collection of Interpretations to present
            to the user
        :returns: The index of the selected item in interpretations.
        """

        if len(interpretations) == 1:
            return 0
        prompt = "Choose an interpretation to use in the search:"
        choices = [Choice(title="all", value=0)]
        i = 1
        for interp in interpretations:
            choices.append(Choice(title=self.interpretation_to_string(interp), value=i))
            i += 1
        return questionary.select(prompt, choices).ask()

    def run_interactive(self) -> None:
        """Run an interactive search loop."""

        previous_search = None
        while True:
            action = questionary.select("What would you like to do?",
                    ["New Search",
                     Choice("Modify Search", disabled=not previous_search),
                     "Edit Settings",
                     "Exit"]).ask()

            if action is None or action == "Exit":
                exit()
            elif action == "New Search":
                previous_search = self.interactive_search()
            elif action == "Modify Search":
                previous_search = self.interactive_search(previous_search)
            elif action == "Edit Settings":
                while True:
                    action = questionary.select("Search Settings",
                            [Choice(title="Uncertainty [{}]".format(self.uncertainty), value=1),
                             Choice(title="Search Mode [{}]".format(self.search_mode.value), value=2),
                             Choice(title="Annotation Mode [{}]".format(self.annotation_mode.value), value=3),
                             Choice(title="Aspirate Mode [{}]".format(self.aspirate_mode.value), value=4),
                             Choice(title="Disjoiner Mode [{}]".format(self.disjoiner_mode.value), value=5),
                             Choice(title="Fix First [{}]".format(self.fix_first), value=6),
                             "Back"]
                            ).ask()

                    if action is None or action == "Back":
                        break
                    elif action == 1:
                        self.change_arg("uncertainty", range(3), display_name="Uncertainty")
                    elif action == 2:
                        self.change_arg("search_mode", list(regen.SearchMode), convert=self.get_enum_value, display_name="Search mode")
                    elif action == 3:
                        self.change_arg("annotation_mode", list(regen.Strictness), convert=self.get_enum_value, display_name="Annotation mode")
                    elif action == 4:
                        self.change_arg("aspirate_mode", list(regen.Strictness), convert=self.get_enum_value, display_name="Aspirate mode")
                    elif action == 5:
                        self.change_arg("disjoiner_mode", list(regen.Strictness), convert=self.get_enum_value, display_name="Disjoiner mode")
                    elif action == 6:
                        self.change_arg("fix_first", [True, False], display_name="Fix First")
                                           
    def get_enum_value(self, enum):
        return enum.value

    def change_arg(self, arg_name, options: Iterable[T], display_name: str=None, convert: Callable[[T], str]=str) -> None:
        """Prompt the user to select the value of a search parameter and set
            the new value.
        
        :param arg_name: The name of the argument to change.
        :param options: A collection of choices to present to the user.
        :param display_name: The title of the prompt.
        :param convert: A function that returns a string given an object of
            type T.
        """

        choices = list()
        for option in options:
            selected = getattr(self, arg_name) == option
            choices.append(Choice(title=convert(option), value=option, checked=selected))
        title = display_name if display_name else "Set " + arg_name
        setting = questionary.select(title, choices).ask()
        if setting is not None:
            setattr(self, arg_name, setting)

    def interactive_search(self, previous: str=None) -> Optional[str]:
        """Run an interactive search.

        :param previous: The previous search performed in this interactive
            session.
        :returns: The search string used.
        """

        search, tree = self.get_grascii_search(previous)
        if search is None:
            return previous
        parses = self.flatten_tree(tree)
        display_interpretations = self.get_unique_interpretations(parses)
        interpretations = list(display_interpretations.values())
        index = self.choose_interpretation(interpretations)
        builder = regen.RegexBuilder(
                uncertainty=self.uncertainty,
                search_mode=self.search_mode, 
                fix_first=self.fix_first, 
                annotation_mode=self.annotation_mode, 
                aspirate_mode=self.aspirate_mode, 
                disjoiner_mode=self.disjoiner_mode
        )
        if index == 0:
            interps = interpretations
        else:
            interps = interpretations[index - 1: index]
        patterns = builder.generate_patterns_map(interps)
        starting_letters = builder.get_starting_letters(interps)
        results = self.perform_search(patterns, starting_letters, metrics.standard)
        count = 0
        display_all = False
        for result in results:
            count += 1
            action = "Next"
            if not display_all: 
                action = questionary.select("Search Results",
                        ["Next",
                         "Display All",
                         "End Search" 
                        ]).ask()
            print(result.strip())
            if action is None or action == "End Search":
                break
            elif action == "Display All":
                display_all = True
            
        print("Results:", count)
        print()
        return search
            
    def get_grascii_search(self, previous: Optional[str]=None) -> Tuple[Optional[str], Optional[Tree]]:
        """Prompt the user for a grascii string.
        
        :param previous: The previous grascii string used in a search.
        :returns: A grascii string and a grascii parse tree.
        """

        while True:
            if previous:
                search = questionary.text("Enter Search:", default=previous).ask()
            else:
                search = questionary.text("Enter Search:").ask()

            if search is None:
                return search, None
            if search == "":
                continue
            search = search.upper()
            result = self.parse_grascii(search)
            if not result:
                previous = search
                continue
            return search, result




import os
import unittest

from lark import Lark, Transformer, Token
from lark.visitors import CollapseAmbiguities

from grascii.grammars import get_grammar
from grascii.dephrase import PhraseFlattener

class TestLessonPhrases(unittest.TestCase):

    # def __init__(self):
        # super.__init__(self)
        # self.parser = Lark.open("phrases.lark",
                # parser="earley")

    # parser = Lark.open("../grammars/phrases.lark",
            # rel_to=__file__,
            # parser="earley")

    # aparser = Lark.open("../grammars/phrases.lark",
            # rel_to=__file__,
            # parser="earley",
            # ambiguity="explicit")

    g = get_grammar("phrases")
    parser = Lark.open(g, parser="earley", ambiguity="resolve")
    aparser = Lark.open(g, parser="earley", ambiguity="explicit")

    trans = PhraseFlattener()

    def _test_lesson(self, number):
        dirname = os.path.dirname(__file__)
        path = os.path.join(dirname, "./phrasing/L" + str(number) + ".txt")
        with open(path, "r") as tests:
            for test in tests:
                phrase, expected = test.strip().split(maxsplit=1)
                with self.subTest(phrase=phrase, expected=expected):
                    tree = self.parser.parse(phrase)
                    tokens = self.trans.transform(tree)
                    tokens = (token.type for token in tokens)
                    parsed = " ".join(tokens)
                    try:
                        self.assertEqual(parsed, expected)
                    except AssertionError as error:
                        if not self.ambiguous_check(phrase, expected):
                            self.fail(msg="failed ambiguity test")


    def ambiguous_check(self, phrase, expected):
        tree = self.aparser.parse(phrase)
        trees = CollapseAmbiguities().transform(tree)
        trans = PhraseFlattener()
        parses = set()
        for t in trees:
            tokens = (token.type for token in trans.transform(t))
            parses.add(" ".join(tokens))
            # print(" ".join(tokens))

        return expected in parses


    def test_lesson1(self):
        self._test_lesson(1)

    def test_lesson2(self):
        self._test_lesson(2)

    def test_lesson3(self):
        self._test_lesson(3)

    def test_lesson4(self):
        self._test_lesson(4)
        
    def test_lesson5(self):
        self._test_lesson(5)

    def test_lesson6(self):
        self._test_lesson(6)

    def test_lesson7(self):
        self._test_lesson(7)

    def test_lesson11a(self):
        self._test_lesson("11a")

    def test_lesson11b(self):
        self._test_lesson("11b")

    def test_lesson11c(self):
        self._test_lesson("11c")

    def test_lesson11d(self):
        self._test_lesson("11d")

    @unittest.skip("fix these later")
    def test_lesson11e(self):
        self._test_lesson("11e")

    def test_lesson11f(self):
        self._test_lesson("11f")

    def test_lesson11i(self):
        self._test_lesson("11i")

    def test_lesson19a(self):
        self._test_lesson("19a")

if __name__ == '__main__':
    unittest.main()



















#!/usr/bin/python

import argparse
from glob import glob
from pathlib import Path
from setuptools import setup, find_packages
from setuptools.command.build_py import build_py

from grascii import dictionary

class CustomBuild(build_py):
    def run(self):
        print(Path("dsrc").glob("*.txt"))
        dictionary.build.build(infiles=Path("dsrc").glob("*.txt"),
                            output=Path("grascii/dictionary/preanniversary"),
                            package=True)
        build_py.run(self)


setup(
    setup_cfg=True,
    cmdclass={
        "build_py": CustomBuild
    }
)




import argparse
import io
from pathlib import Path
from typing import TextIO
import sys

from pkg_resources import resource_stream, resource_exists

from grascii.dictionary import list as list_dict, install, uninstall, build
from grascii.dictionary.install import DICTIONARY_PATH

description = "Create and manage Grascii dictionaries"

def build_argparser(argparser: argparse.ArgumentParser) -> None:
    argparser.set_defaults(func=list_dict.cli_list)
    subparsers = argparser.add_subparsers(title="subcommands")

    build_parser = subparsers.add_parser("build", 
            description=build.description, 
            help=build.description,
            aliases=["b"])
    build.build_argparser(build_parser)
    build_parser.set_defaults(func=build.cli_build)

    install_parser = subparsers.add_parser("install",
            description=install.description,
            help=install.description,
            aliases=["i"])
    install.build_argparser(install_parser)
    install_parser.set_defaults(func=install.cli_install)

    uninstall_parser = subparsers.add_parser("uninstall",
            description=uninstall.description,
            help=uninstall.description,
            aliases=["u"])
    uninstall.build_argparser(uninstall_parser)
    uninstall_parser.set_defaults(func=uninstall.cli_uninstall)

    list_parser = subparsers.add_parser("list",
            description=list_dict.description,
            help=list_dict.description,
            aliases=["l"])
    list_dict.build_argparser(list_parser)
    list_parser.set_defaults(func=list_dict.cli_list)

def get_dict_file(dictionary: str, name: str) -> TextIO:
    if dictionary[0] == ":":
        if resource_exists("grascii.dictionary", dictionary[1:]):
            module = "grascii.dictionary." + dictionary[1:]
            return io.TextIOWrapper(resource_stream(module, name),
                    encoding="utf-8")
        directory = DICTIONARY_PATH.joinpath(dictionary[1:])
        if not directory.exists():
            print(dictionary, "does not exist.", file=sys.stderr)
        return directory.joinpath(name).open()
    return Path(dictionary, name).open()





#!/usr/bin/env python

import sys
import argparse
import os

from grascii import search, config, dictionary

def main() -> None:
    argparser = argparse.ArgumentParser(prog="grascii")
    subparsers = argparser.add_subparsers(title="subcommands")   
    argparser.set_defaults(func=None)

    search_parser = subparsers.add_parser("search", 
            description=search.description, 
            help=search.description,
            aliases=["s"])
    search.build_argparser(search_parser)
    search_parser.set_defaults(func=search.cli_search)

    dictionary_parser = subparsers.add_parser("dictionary", 
            description=dictionary.description, 
            help=dictionary.description,
            aliases=["d"])
    dictionary.build_argparser(dictionary_parser)

    config_parser = subparsers.add_parser("config", 
            description=config.description, 
            help=config.description,
            aliases=["cf"])
    config.build_argparser(config_parser)
    config_parser.set_defaults(func=config.cli_config)

    args = argparser.parse_args(sys.argv[1:])

    if args.func:
        args.func(args)
    else:
        argparser.print_help()


if __name__ == "__main__":
    main()




"""Contains functions for working with a grascii configuration file.
Acts as the entry point for the grascii config command.

This can be invoked as a standalone program:
$ python -m grascii.config --help
"""

import argparse
from pathlib import Path, PurePath
from shutil import copyfile
import sys

from grascii.appdirs import user_config_dir

APP_NAME = "grascii"
CONF_DIRECTORY = user_config_dir(APP_NAME)
CONF_FILE_NAME = APP_NAME + ".conf"

description = "Manage Grascii configuration"

def build_argparser(argparser: argparse.ArgumentParser) -> None: 
    """Configure an ArgumentParser parser to parse the config command-line
    options
    
    :parse argparser: A fresh ArgumentParser to configure.
    """

    group = argparser.add_mutually_exclusive_group(required=True)
    group.add_argument("--init", action="store_true",
            help="Create a configuration file.")
    group.add_argument("-D", "--delete", action="store_true",
            help="Delete an existing configuration file.")
    group.add_argument("-w", "--where", action="store_true",
            help="Print the path to the configuration file and exit.")
    argparser.add_argument("-f", "--force", action="store_true",
            help="Allow overwriting of an existing configuration file.")

def config_exists() -> bool:
    """Check whether the user configuration file exists.
    
    :returns: True if the configuration file exists.
    """

    return Path(get_config_file_path()).exists()

def get_config_file_path() -> PurePath:
    """Get a path object representating the location of the configuration
    file.
    
    :returns: A path object.
    """

    return PurePath(CONF_DIRECTORY, CONF_FILE_NAME)

def create_config() -> None:
    """Create a configuration file with the default settings."""

    Path(CONF_DIRECTORY).mkdir(parents=True, exist_ok=True)
    src = Path(__file__).with_name("defaults.conf")
    copyfile(src, get_config_file_path())

def delete_config() -> None:
    """Delete a configuration file."""

    Path(get_config_file_path()).unlink()


def cli_config(args: argparse.Namespace) -> None:
    """Run a search using arguments parsed from the command line.
    
    :param args: A namespace of parsed arguments.
    """

    if args.where:
        if config_exists():
            print(get_config_file_path())
        else:
            print("Configuration file does not exist. Use --init to create one.", file=sys.stderr)
    elif args.init:
        if config_exists() and not args.force:
            print("Configuration file already exists.", file=sys.stderr)
            print("If you would like to overwrite it with the default", 
                  "configuration, run again with --force.", file=sys.stderr)
            return
        create_config()
        print("Configuration file created at", get_config_file_path())
    elif args.delete:
        if not config_exists():
            print("Configuration file does not exist.", file=sys.stderr)
            return
        if not args.force:
            print("Are you sure you want to delete the configuration file?",
                  "If so, run with --force.", file=sys.stderr)
            return
        delete_config()
        print("Removed", get_config_file_path())

def main() -> None:
    """Run the config command using arguments from sys.argv."""

    argparser = argparse.ArgumentParser(description)
    build_argparser(argparser)
    args = argparser.parse_args(sys.argv[1:])
    cli_config(args)

if __name__ == "__main__":
    main()





